import struct
import typing
from dataclasses import dataclass
from functools import cache

from .utils import get_usable_fields
from .types import SerializableType, DeserializedT, SimpleSerializableType, RecursiveType

if typing.TYPE_CHECKING:
    from .deserializer import Deserializer
    from .serializer import Serializer


@dataclass
class Plan:
    pattern: str
    total_size: int
    fields: list[str]
    serializable_type: typing.Optional[SerializableType] = None
    type_: typing.Optional[type[DeserializedT]] = None

    def read_using_plan(self, deserializer: "Deserializer", instance, parent_custom_type=None):
        if self.type_ is not None:
            setattr(instance, self.fields[0],
                    deserializer.deserialize(self.type_, parent_custom_type=parent_custom_type))
        elif self.serializable_type is not None:
            setattr(instance, self.fields[0],
                    self.serializable_type.load_from_bytes(deserializer, instance,
                                                           parent_custom_type=parent_custom_type))
        else:
            vals = struct.unpack(self.pattern, deserializer.reader.read_bytes(self.total_size))
            for field, value in zip(self.fields, vals):
                setattr(instance, field, value)

    def write_using_plan(self, serializer: "Serializer", instance, parent_custom_type=None):
        if self.type_ is not None:
            serializer.serialize(getattr(instance, self.fields[0]), parent_custom_type=parent_custom_type)
        elif self.serializable_type is not None:
            self.serializable_type.write_from_value(serializer, getattr(instance, self.fields[0]), parent_instance=instance,
                                                    parent_custom_type=parent_custom_type)
        else:
            fields_to_write = [getattr(instance, field) for field in self.fields]
            data_to_write = struct.pack(self.pattern, *fields_to_write)
            serializer.writer.write_bytes(data_to_write)


@cache
def generate_deserializing_plans(type_: type, first=None, last=None):
    plans = []
    plan = None

    for key, val in get_usable_fields(type_, first, last):
        if isinstance(val, SimpleSerializableType):
            if plan is None:
                plan = Plan("", 0, [])

            plan.pattern += val.STRUCT_PATTERN[
                            1:] if plan.pattern else val.STRUCT_PATTERN  # todo: do something with the endinianity
            plan.total_size += val.SIZE
            plan.fields.append(key)
        else:
            if plan is not None:
                plans.append(plan)
                plan = None

            if isinstance(val, SerializableType):
                if isinstance(val, RecursiveType):
                    val = type_
                plans.append(Plan(pattern="", total_size=0, fields=[key], serializable_type=val))
                continue

            if isinstance(val, RecursiveType):
                val = type_
            plans.append(Plan(pattern="", total_size=0, fields=[key], type_=val))

    if plan is not None:
        plans.append(plan)

    return plans



import typing


class Writer(typing.Protocol):
    def write_bytes(self, data: bytes):
        raise NotImplementedError


class FileWriter:
    def __init__(self, file_path=None, file=None):
        self.file = file or open(file_path, "wb")

    def write_bytes(self, data: bytes):
        self.file.write(data)


class SizeCalculatorWriter:
    def __init__(self):
        self.current_size = 0

    def write_bytes(self, data: bytes):
        self.current_size += len(data)


class BufferWriter:
    def __init__(self):
        self.buffer = bytearray()

    def write_bytes(self, data: bytes):
        self.buffer.extend(data)



import binpi


class BasicStructure:
    le_int = binpi.LEInt()
    le_uint = binpi.LEUInt()
    le_short = binpi.LEShort()
    le_ushort = binpi.LEUShort()
    le_byte = binpi.LEByte()
    le_ubyte = binpi.LEUByte()
    le_float = binpi.LEFloat()
    le_double = binpi.LEDouble()

    be_int = binpi.BEInt()
    be_uint = binpi.BEUInt()
    be_short = binpi.BEShort()
    be_ushort = binpi.BEUShort()
    be_byte = binpi.BEByte()
    be_ubyte = binpi.BEUByte()
    be_float = binpi.BEFloat()
    be_double = binpi.BEDouble()



import typing


class Reader(typing.Protocol):
    def read_bytes(self, n: int):
        raise NotImplementedError


class FileReader:
    def __init__(self, file_path=None, file=None):
        self.file = file or open(file_path, "rb")

    def read_bytes(self, n: int):
        return self.file.read(n)


class BufferReader:
    offset: int
    data: bytes

    def __init__(self, data: bytes, offset: int = 0):
        self.data = data
        self.offset = offset

    def read_bytes(self, n: int):
        data = self.data[self.offset: self.offset + n]
        self.offset += n
        return data



import unittest
import binpi

from .shared import BasicStructure


class SerializeTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        data = BasicStructure()
        data.le_int = 100
        data.le_uint = 100
        data.le_short = 100
        data.le_ushort = 100
        data.le_byte = 100
        data.le_ubyte = 100
        data.le_float = 100
        data.le_double = 100
        data.be_int = 100
        data.be_uint = 100
        data.be_short = 100
        data.be_ushort = 100
        data.be_byte = 100
        data.be_ubyte = 100
        data.be_float = 100
        data.be_double = 100
        self.data = data

    def test_size(self):
        self.assertEqual(binpi.get_serialized_size(self.data), 52)

    def test_serialize_all(self):
        writer = binpi.BufferWriter()
        binpi.serialize(self.data, writer=writer)
        self.assertEqual(len(writer.buffer), 52)
        # todo: assert the data

    def test_serialize_half(self):
        writer = binpi.BufferWriter()
        binpi.serialize(self.data, writer=writer, last="le_double")
        self.assertEqual(len(writer.buffer), 26)
        # todo: assert the data

    def test_serialize_second_half(self):
        writer = binpi.BufferWriter()
        binpi.serialize(self.data, writer=writer, first="be_int", last="be_double")
        self.assertEqual(len(writer.buffer), 26)
        # todo: assert the data


if __name__ == '__main__':
    unittest.main()



from .plans import generate_deserializing_plans
from .types import RecursiveType
from .writer import Writer


class Serializer:
    def __init__(self, writer: Writer = None):
        self.writer = writer

    def serialize(self, value, first=None, last=None, parent_custom_type=None):
        value_type_ = type(value)

        if isinstance(value_type_, RecursiveType):
            value_type_ = parent_custom_type

        plans = generate_deserializing_plans(value_type_, first=first, last=last)
        for plan in plans:
            plan.write_using_plan(self, value, parent_custom_type=value_type_)



from .plans import generate_deserializing_plans
from .types import RecursiveType
from .types import _WrapType
from .reader import Reader, BufferReader


class Deserializer:
    def __init__(self, reader: Reader = None, bytes: bytes = None):
        self.reader = reader or BufferReader(bytes)

    def deserialize(self, class_, parent_custom_type=None, first=None, last=None):
        if isinstance(class_, _WrapType):
            class_ = class_.type

        if isinstance(class_, RecursiveType):
            class_ = parent_custom_type

        result = class_()

        plans = generate_deserializing_plans(class_, first=first, last=last)
        for plan in plans:
            plan.read_using_plan(self, result, parent_custom_type=class_)

        return result



import typing
from typing import Callable

if typing.TYPE_CHECKING:
    from .serializer import Serializer
    from .deserializer import Deserializer

DeserializedT = typing.TypeVar("DeserializedT")


class SimpleSerializableType:
    STRUCT_PATTERN: str = ""
    SIZE: int = -1

    def get_SIZE(self):
        return self.SIZE

    def get_STRUCT_PATTERN(self):
        return self.STRUCT_PATTERN


class SerializableType:
    def load_from_bytes(self, deserializer: "Deserializer", instance, *args, **kwargs):
        raise NotImplementedError

    def write_from_value(self, serializer: "Serializer", value, parent_instance, *args, **kwargs):
        raise NotImplementedError


class Skip():
    ...


def create_simple_number_class(format: str, size: int) -> type[int]:
    class _inner(SimpleSerializableType):
        STRUCT_PATTERN = format
        SIZE = size

    return _inner  # type: ignore


def create_simple_float_class(format: str, size: int) -> type[float]:
    class _inner(SimpleSerializableType):
        STRUCT_PATTERN = format
        SIZE = size

    return _inner  # type: ignore


LEInt: Callable[..., int] = create_simple_number_class("<i", 4)
LEUInt: Callable[..., int] = create_simple_number_class("<I", 4)
LEShort: Callable[..., int] = create_simple_number_class("<h", 2)
LEUShort: Callable[..., int] = create_simple_number_class("<H", 2)
LEByte: Callable[..., int] = create_simple_number_class("<b", 1)
LEUByte: Callable[..., int] = create_simple_number_class("<B", 1)
LEFloat: Callable[..., float] = create_simple_float_class("<f", 4)
LEDouble: Callable[..., float] = create_simple_float_class("<d", 8)

BEInt: Callable[..., int] = create_simple_number_class(">i", 4)
BEUInt: Callable[..., int] = create_simple_number_class(">I", 4)
BEShort: Callable[..., int] = create_simple_number_class(">h", 2)
BEUShort: Callable[..., int] = create_simple_number_class(">H", 2)
BEByte: Callable[..., int] = create_simple_number_class(">b", 1)
BEUByte: Callable[..., int] = create_simple_number_class(">B", 1)
BEFloat: Callable[..., float] = create_simple_float_class(">f", 4)
BEDouble: Callable[..., float] = create_simple_float_class(">d", 8)


class _Boolean(SimpleSerializableType):
    STRUCT_PATTERN = "<?"
    SIZE = 1


Boolean: Callable[..., bool] = _Boolean  # type: ignore


class _WrapType:
    def __init__(self, type_):
        self.type = type_


WrapTypeT = typing.TypeVar('WrapTypeT')


def WrapType(type_: WrapTypeT) -> WrapTypeT:
    return _WrapType(type_)


class RecursiveType:
    def __init__(self):
        self.type = None



from .reader import *
from .writer import *
from .types import *
from .utils import *
from .deserializer import *
from .serializer import *
from .list import *
from .into_type_instance import *


def deserialize(class_: type[DeserializedT], reader: Reader = None, first=None, last=None, bytes=None,
                parent_custom_type=None) -> DeserializedT:
    return Deserializer(reader=reader, bytes=bytes).deserialize(class_=class_, first=first, last=last,
                                                                parent_custom_type=parent_custom_type)


def serialize(value, writer: Writer, first=None, last=None, parent_custom_type=None):
    return Serializer(writer=writer).serialize(value=value, first=first, last=last,
                                               parent_custom_type=parent_custom_type)


def get_serialized_size(value, first=None, last=None) -> int:
    """ Useful for archive structures that stores headers and data separately and uses offsets for reading the data """
    """ NOTE: this function is quite expensive to call on big data structures """
    writer = SizeCalculatorWriter()
    serialize(value, writer, first=first, last=last)
    return writer.current_size



from binpi import DeserializedT


def _get_value(type, value):
    if isinstance(value, list):
        item_type = type.type
        return [_get_value(item_type, item) for item in value]
    elif isinstance(value, dict):
        instance = type()
        for key, val in value.items():
            setattr(instance, key, _get_value(type.__dict__.get(key), val))
        return instance
    else:
        return value


def into_type_instance(type: type[DeserializedT], data) -> DeserializedT:
    return _get_value(type, data)



from functools import cache
from inspect import ismethod, isfunction

from .types import Skip


@cache
def get_usable_fields(class_, first=None, last=None):
    pairs = [(attr, val) for attr, val in class_.__dict__.items() if
             not attr.startswith("__") and not isinstance(val, Skip) and not callable(val) and not ismethod(val) and not isfunction(val)]

    first_index, last_index = 0, len(pairs)
    if first is not None:
        first_index = next(i for i in range(len(pairs)) if pairs[i][0] == first)
    if last is not None:
        last_index = next(i for i in range(len(pairs)) if pairs[i][0] == last)

    return pairs[first_index: min(last_index + 1, len(pairs))]



from setuptools import setup, find_packages

setup(
      name='binpi',
      version='0.1.14',
      description='Simple library for de/serializing binary data',
      readme = "README.md",
      long_description=open("README.md", "r").read(),
      long_description_content_type="text/markdown",
      url='https://github.com/romansvozil/binpi',
      author='Roman Svozil',
      license='MIT',
      packages=find_packages(),
)


import struct
import typing

if typing.TYPE_CHECKING:
    from .deserializer import Deserializer
    from .serializer import Serializer

from .types import SerializableType, DeserializedT, SimpleSerializableType, LEUByte, BEUByte


class _List(SerializableType):
    size: int | str | typing.Callable
    type: type[DeserializedT]

    def __init__(self, type_, size: int | str | typing.Callable, **kwargs):
        super().__init__(**kwargs)
        self.size = size
        self.type = type_

    def load_from_bytes(self, deserializer: "Deserializer", instance, *args, **kwargs):
        result = []

        if isinstance(self.type, SimpleSerializableType):
            size = self.get_size(instance)
            if size == 0:
                return result
            pattern = self.type.get_STRUCT_PATTERN()
            full_pattern = pattern[0] + str(size) + pattern[1:] # todo: again the endianity is annoying
            bytes = deserializer.reader.read_bytes(size * self.type.get_SIZE())
            return list(struct.unpack(full_pattern, bytes))

        for index in range(self.get_size(instance)):
            if isinstance(self.type, SerializableType):
                result.append(self.type.load_from_bytes(deserializer, instance,
                                                        parent_custom_type=kwargs.get("parent_custom_type", None)))
            else:
                result.append(deserializer.deserialize(self.type, parent_custom_type=kwargs.get("parent_custom_type", None)))

        return result

    def write_from_value(self, serializer: "Serializer", value, parent_instance, *args, **kwargs):
        if isinstance(self.type, SimpleSerializableType):
            size = len(value)
            if size == 0:
                return
            pattern = self.type.get_STRUCT_PATTERN()
            full_pattern = pattern[0] + str(size) + pattern[1:] # todo: the endinianity is annoying

            serializer.writer.write_bytes(struct.pack(full_pattern, *value))
            return

        for val in value:
            if isinstance(self.type, SerializableType):
                self.type.write_from_value(serializer, val, parent_instance)
            else:
                serializer.serialize(val, *args, **kwargs)

    def get_size(self, instance):
        return self.size \
            if type(self.size) == int \
            else self.size(instance) \
            if callable(self.size) \
            else getattr(instance, self.size)


class _String(_List):
    def __init__(self, type_=LEUByte(), size: int | str | typing.Callable = 0, encoding: str = "utf8", **kwargs):
        super().__init__(type_=type_, size=size, **kwargs)
        self.encoding = encoding

    def load_from_bytes(self, deserializer: "Deserializer", instance, *args, **kwargs):
        result = super().load_from_bytes(deserializer, instance, *args, **kwargs)
        return bytes(result).decode(self.encoding)

    def write_from_value(self, serializer: "Serializer", value: str, parent_instance, *args, **kwargs):
        return super().write_from_value(serializer, value.encode(self.encoding), parent_instance, *args, **kwargs)


ListItemT = typing.TypeVar("ListItemT")


def List(type_: type[ListItemT] | ListItemT, size: int | str | typing.Callable, *args, **kwargs) -> typing.List[ListItemT]:
    # quite hacky way of doing this, but it is what it is
    return _List(type_, size, *args, **kwargs)  # type: ignore


def String(type_: type = BEUByte(), size: int | str | typing.Callable = 0, encoding: str = "utf8", *args, **kwargs) -> str:
    # quite hacky way of doing this, but it is what it is
    return _String(type_, size, encoding, *args, **kwargs)  # type: ignore